<?php

namespace InSegment\ApiCore\Models;

use Illuminate\Support\Facades\DB;
use Illuminate\Support\Facades\Schema;
use Illuminate\Database\Schema\Blueprint;

class UUIDGeneration extends SupplementaryTable
{
    /**
     * Run the migrations.
     *
     * @param array $enabledTables
     * @param bool $inMemory
     * @param string|null $uid
     * @return void
     */
    public static function establish(array $enabledTables, bool $inMemory, $uid = null)
    {
        Schema::create((new static([], $uid))->getTable(), function (Blueprint $table) use ($enabledTables, $inMemory) {
            if ($inMemory) {
                $table->engine = 'MEMORY';
            }
            
            $table->unsignedInteger('id')->nullable();
            $table->unsignedBigInteger('uuid');
            $table->enum('table', $enabledTables);
            $table->primary(['uuid'], 'uuid_generation_primary');
            $table->index(['table'], 'uuid_generation_table_index');
        });
    }
    
    /**
     * Generate identifiers for tables with auto-increment primary keys and update the record of UUID mapping
     * 
     * @param array $allowedUuidables [
     *     $table => $keyName,
     *     ...
     * ]
     * @param string $generationsTbl
     * @return null
     */
    public static function generateIdentifiers($allowedUuidables, $generationsTbl)
    {
        list($gensDb, $gensTableName) = explode('.', $generationsTbl);
        
        foreach ($allowedUuidables as $table => $keyName) {
            DB::statement("SELECT * FROM `{$table}` WHERE `{$keyName}` > (SELECT MAX(`{$keyName}`) FROM `{$table}`) FOR UPDATE");
            DB::statement("SET @ai = (SELECT `AUTO_INCREMENT` FROM `information_schema`.`tables` WHERE `table_name` = '{$table}' AND `table_schema` = DATABASE())");
            DB::statement("UPDATE `{$gensDb}`.`{$gensTableName}` SET `id` = (@ai:= @ai + 1) - 1 WHERE `table` = '{$table}'");
        }
    }
    
    /**
     * Update statement replacing UUIDs with auto-increment primary keys
     * 
     * @param string $tempTable
     * @param string $generationsTbl
     * @param string $key
     * @return string
     */
    public static function getUUIDUpdateStatement(string $tempTable, string $generationsTbl, string $key)
    {
        list($databaseTemp, $tempTableName) = explode('.', $tempTable);
        list($databaseGens, $gensTableName) = explode('.', $generationsTbl);
        return
            "UPDATE `{$databaseTemp}`.`{$tempTableName}` as `temp_table`"
            . " INNER JOIN `{$databaseGens}`.`{$gensTableName}` as `gens_table` ON `gens_table`.`uuid` = `temp_table`.`{$key}`"
            . " SET `temp_table`.`{$key}` = `gens_table`.`id`;";
    }
    
    /**
     * Get statement generating amount if ids for table and key
     * 
     * @param string $table
     * @param int $rows
     * @param string|null $uid
     * @return null
     */
    public static function generateForTable(string $table, int $rows, $uid = null)
    {
        list($schema, $myTable) = explode('.', (new static([], $uid))->getTable());
        $nRows = implode(', ', array_fill(0, $rows, '()'));
        
        $pdo = DB::connection()->getPdo();
        $emulatePrepares = $pdo->getAttribute(\PDO::ATTR_EMULATE_PREPARES);
        
        if (!$emulatePrepares) {
            $pdo->setAttribute(\PDO::ATTR_EMULATE_PREPARES, true);
        }
        
        DB::statement(
<<<SQL
CREATE TRIGGER `{$schema}`.`trigger_{$myTable}` BEFORE INSERT ON `{$schema}`.`{$myTable}` FOR EACH ROW SET NEW.`uuid` = UUID_SHORT(), NEW.`table` = '{$table}';
INSERT INTO `{$schema}`.`{$myTable}` () VALUES {$nRows};
DROP TRIGGER `{$schema}`.`trigger_{$myTable}`;
SQL
        );

        if (!$emulatePrepares) {
            $pdo->setAttribute(\PDO::ATTR_EMULATE_PREPARES, false);
        }
    }
}
