<?php

namespace InSegment\ApiCore\Services;

use InSegment\ApiCore\Exceptions\ApiTransactorException;
use InSegment\ApiCore\Services\ParseCreateTable;

use Illuminate\Support\Facades\DB;
use Illuminate\Support\Facades\Cache;

use InSegment\ApiCore\Interfaces\SliceManagementInterface as SliceManager;

class SliceMerger
{
    // TODO: refactor Transactor to use these constants
    const TEMP_TABLE_TYPE_INSERT = 'insert';
    const TEMP_TABLE_TYPE_INSERT_IGNORE = 'place';
    const TEMP_TABLE_TYPE_IODKU = 'update';
    const TEMP_TABLE_TYPE_REPLACE = 'replace';
    
    const DEFAULT_TEMP_TABLE_TYPES = [
        self::TEMP_TABLE_TYPE_INSERT,
        self::TEMP_TABLE_TYPE_IODKU,
    ];
    
    const CLEARABLE_TEMP_TABLE_TYPES = [
        self::TEMP_TABLE_TYPE_INSERT,
        self::TEMP_TABLE_TYPE_INSERT_IGNORE,
        self::TEMP_TABLE_TYPE_IODKU,
        self::TEMP_TABLE_TYPE_REPLACE,
    ];
    
    /**
     * Show create table parser
     * 
     * @var \InSegment\ApiCore\Services\ParseCreateTable 
     */
    private $showCreateParser;
    
    /**
     * Slice manager
     *
     * @var \InSegment\ApiCore\Interfaces\SliceManagementInterface
     */
    private $sliceManager;
    
    /**
     * Temporary table types
     * 
     * @var array
     */
    private $tempTableTypes;
    
    /**
     * Store for columns listings for tables
     * 
     * @var array [
     *     $table => string[] $columns,
     *     ...
     * ]
     */
    private $columnStore = [];
    
    /**
     * Store for ODKU statement parts
     * 
     * @var array [
     *     $table => string $odku,
     *     ...
     * ]
     */
    private $ODKUStore = [];
    
    /**
     * Cache prefix for slice merger
     *
     * @var string
     */
    private $cachePrefix;
    
    /**
     * Constructor
     * 
     * @param \InSegment\ApiCore\Interfaces\SliceManagementInterface $sliceManager
     * @param \InSegment\ApiCore\Services\ParseCreateTable $showCreateParser
     * @param array $tempTableTypes
     */
    public function __construct(SliceManager $sliceManager, ParseCreateTable $showCreateParser, array $tempTableTypes = self::DEFAULT_TEMP_TABLE_TYPES)
    {
        $this->sliceManager = $sliceManager;
        $this->showCreateParser = $showCreateParser;
        $this->tempTableTypes = $tempTableTypes;
        $this->cachePrefix = config('api_core.cache_prefix', 'insegment.api-core.') . 'slice-merger';
    }
    
    /**
     * Create temporary table
     * 
     * @param string $table
     * @param string[] $uuidableTableMentions
     * @param bool $isUuidable
     * @param bool $allowInMemory
     * @return [
     *     string $type => string $tempTableOfType
     * ],
     * @throws \InSegment\ApiCore\Exceptions\ApiTransactorException
     */
    public function createTempTable($table, $uuidableTableMentions, $isUuidable, $allowInMemory = false)
    {
        $createStatementSplit = $this->showCreateParser->getCreateForTemp($table, $uuidableTableMentions, $isUuidable, $allowInMemory);
        $tempTables = [];
        
        // first, create temporary tables
        foreach ($this->tempTableTypes as $type) {
            list($tempTableDb, $tempTable) = explode('.', $tempTables[$type] = $this->prefixTableForTransaction($table, $type));
            $createStatementSplit[0] = "CREATE TABLE `{$tempTableDb}`.`{$tempTable}` (";
            if (!DB::statement(implode("\n", $createStatementSplit))) {
                throw (new ApiTransactorException(ApiTransactorException::CODE_CREATE_TABLE_FAIL))
                    ->compile($type, $table);
            }
        }
        
        return $tempTables;
    }
    
    /**
     * Drop temporary table
     * 
     * @param array $tempTableDef [
     *     string $type => string $tempTableOfType
     * ]
     * @return null
     */
    public function dropTempTable(array $tempTableDef)
    {
        foreach ($this->tempTableTypes as $type) {
            list ($tempTableDb, $tempTableName) = explode('.', $tempTableDef[$type]);
            DB::statement("DROP TABLE IF EXISTS `{$tempTableDb}`.`{$tempTableName}`");
        }
    }
    
    /**
     * Merge temporary tables into real ones
     * 
     * @param array $options [
     *     'tempTables' => &[ // enumeration of temporary tables
     *         string $tableName => [
     *             string $type => string $tempTableOfType,
     *             ...
     *         ],
     *         ...
     *     ],
     *     'allowedUuidables' => [ // all tables with auto-incrementing primary keys which are listed in uuidables and transaction
     *         string $tableName => string $keyName,
     *         ...
     *     ],
     *     'uuidableMentions' => [ // tables to the keys which are mentioned in some uuidable tables and thus may have updated key
     *         string $tableName => string|string[] $keys,
     *         ...
     *     ],
     *     'deletionEnabledTables' => array|null,
     *     'postProcessing' => callable function (\InSegment\ApiCore\Interfaces\SliceManagementInterface $sliceManager),
     *     'supplementaries' => [
     *         'generations' => string $generationsModelClass, // generations is the Model for table which contains new identifiers
     *         'deletes' => string|null $deletesModelClass, // this Model's table lists records identifiers to be deleted from real tables
     *         'logs' => string|null $logsModelClass, // this Model's table with transacted records identifiers logged for use in post-processing
     *     ]
     * ]
     * @return null
     * @throws \InSegment\ApiCore\Exceptions\ApiTransactorException
     */
    public function mergeTempTables(array $options)
    {
        $currentTable = null;
        $currentOp = null;
        
        try {
            DB::transaction(function () use (&$options, &$currentTable, &$currentOp) {
                $uid = $this->sliceManager->getUID();
                $generationsModelClass = $options['supplementaries']['generations'];
                $deletesModelClass = array_get($options['supplementaries'], 'deletes');
                $logsModelClass = array_get($options['supplementaries'], 'logs');
                
                $generationsTbl = (new $generationsModelClass([], $uid))->getTable();
                $generationsModelClass::generateIdentifiers($options['allowedUuidables'], $generationsTbl);
                $deletesTable = $deletesModelClass ? (new $deletesModelClass([], $uid))->getTable() : null;
                $logsTable = $logsModelClass ? (new $logsModelClass([], $uid))->getTable() : null;
                
                foreach ($options['uuidableMentions'] as $table => $keys) {
                    foreach (array_wrap($keys) as $key) {
                        foreach ($this->tempTableTypes as $type) {
                            if (!isset($options['tempTables'][$table][$type])) {
                                continue;
                            }

                            DB::statement($generationsModelClass::getUUIDUpdateStatement($options['tempTables'][$table][$type], $generationsTbl, $key));
                        }
                    }
                }
                
                isset($deletesTable) && DB::statement($generationsModelClass::getUUIDUpdateStatement($deletesTable, $generationsTbl, 'key'));
                isset($logsTable) && DB::statement($generationsModelClass::getUUIDUpdateStatement($logsTable, $generationsTbl, 'key'));
                
                if (isset($deletesTable) && count(array_get($options, 'deletionEnabledTables', []))) {
                    $currentOp = 'groupsForDelete';
                    $groupsForDelete = $deletesModelClass::groupBy('table')->pluck('key_name', 'table');
                    list($deletesDb, $deletesTableName) = explode('.', $deletesTable);
                    
                    foreach ($groupsForDelete as $table => $keyName) {
                        $currentOp = 'delete';
                        
                        DB::statement(
                            "DELETE `{$table}` FROM `{$table}`"
                            . " INNER JOIN `{$deletesDb}`.`{$deletesTableName}` as `deletes`"
                                . " ON `deletes`.`table` = '{$table}'"
                                . " AND `deletes`.`key` = `{$table}`.`{$keyName}`"
                        );
                    }
                }
                
                foreach ($options['tempTables'] as $table => $tempTable) {
                    $currentTable = $table;
                    foreach ($this->tempTableTypes as $type) {
                        $currentOp = $type;
                        $this->doMoveTable($table, $tempTable[$type], $type);
                    }
                }
                
                $currentOp = 'postProcess';
                DB::statement("SET FOREIGN_KEY_CHECKS=1");
                $options['postProcessing']($this->sliceManager);
                DB::statement("SET FOREIGN_KEY_CHECKS=0");
            });
        } catch (\Throwable $exception) {
            if ($currentOp === 'postProcess') {
                throw (new ApiTransactorException(ApiTransactorException::CODE_POST_PROCESS_FAILED, $exception))->compile();
            } else {
                throw (new ApiTransactorException(ApiTransactorException::CODE_FAIL_UNDONE, $exception))
                    ->compile($currentOp, $currentTable);
            }
        }
    }
    
    /**
     * Get a list of table names (not necessarily created ones!) for temporary tables
     * 
     * @param string[] $tableNames
     * @return [
     *     string $tableName => [
     *         string $type => string $tempTableOfType,
     *         ...
     *     ],
     *     ...
     * ]
     */
    public function listTemporaryTableNames($tableNames)
    {
        $result = [];
        
        foreach ($tableNames as $tableName) {
            foreach ($this->tempTableTypes as $type) {
                $tempTableOfType = $this->prefixTableForTransaction($tableName, $type);
                $result[$tableName][$type] = $tempTableOfType;
            }
        }
        
        return $result;
    }
    
    /**
     * REPLACE statement from transaction table
     * 
     * @param string $table
     * @param string $moveTable
     * @param srtring $type
     * @param int $shouldBeReplaced
     * @return null
     * @throw \InSegment\ApiCore\Exceptions\ApiTransactorException
     */
    private function doMoveTable($table, $moveTable, $type)
    {
        list($database, $moveTableName) = explode('.', $moveTable);
        $odku = '';
        
        switch ($type) {
            case self::TEMP_TABLE_TYPE_INSERT:
                $op = 'INSERT';
                $estimate = $this->sliceManager->getNewCount($table);
            break;
            case self::TEMP_TABLE_TYPE_INSERT_IGNORE:
                $op = 'INSERT IGNORE';
                $estimate = null;
            break;
            case self::TEMP_TABLE_TYPE_IODKU:
                $op = 'INSERT';
                $odku = ' ON DUPLICATE KEY UPDATE ' . $this->getODKUForTable($table); // TODO: handle better specific changed values for ODKU
                $estimate = null;
            break;
            case self::TEMP_TABLE_TYPE_REPLACE:
                $op = 'REPLACE';
                $estimate = 2 * ($this->sliceManager->getWrittenCount($table) - $this->sliceManager->getNewCount($table));
            break;
            default: throw (new ApiTransactorException(ApiTransactorException::CODE_UNKNOWN_TYPE_OF_MOVE_OPERATION))->compile($type);
        }
        
        // do not risk and waste time on nothing
        if ($estimate === 0) {
            return;
        }

        $affected = DB::affectingStatement(
            "{$op} INTO `{$table}` SELECT * FROM `{$database}`.`{$moveTableName}`{$odku}"
        );

        if (isset($estimate) && $affected !== $estimate) {
            throw (new ApiTransactorException(ApiTransactorException::CODE_UNEXPECTED_CASUALTIES))
                ->compile($type, $moveTableName, $affected, $op, $estimate);
        }
    }
    
    /**
     * Get ON DUPLICATE KEY UPDATE column listing for table
     * 
     * @param string $table
     * @return string
     */
    public function getTableColumns($table)
    {
        if (!isset($this->columnStore[$table])) {
            $cacheKey = "{$this->cachePrefix}.columns.{$table}";
            $this->columnStore[$table] = Cache::rememberForever($cacheKey, function () use ($table) {
                return DB::getSchemaBuilder()->getColumnListing($table);
            });
        }
        
        return $this->columnStore[$table];
    }
            
    /**
     * Get ON DUPLICATE KEY UPDATE column listing for table
     * 
     * @param string $table
     * @return string
     */
    private function getODKUForTable($table)
    {
        if (!isset($this->ODKUStore[$table])) {
            $cacheKey = "{$this->cachePrefix}.odku.{$table}";
            $this->ODKUStore[$table] = Cache::rememberForever($cacheKey, function () use ($table) {
                $listing = $this->getTableColumns($table);
                return implode(',', array_map(function ($column) {
                    return "`{$column}` = VALUES(`{$column}`)";
                }, $listing));
            });
        }
        
        return $this->ODKUStore[$table];
    }
            
    /**
     * Prefixes table name for transaction table
     * 
     * @param string $tableName
     * @param string $type
     * @return string
     */
    private function prefixTableForTransaction(string $tableName, string $type): string
    {
        $database = $this->sliceManager->getSchema();
        $uid = $this->sliceManager->getUID();
        $ret = "{$tableName}_{$type}_{$uid}";
        
        // limit of table name length is 64
        if (strlen($ret) > 64) {
            return substr($ret, 0, 64);
        }
        
        return "{$database}.{$ret}";
    }
    
}
